import os
import sqlite3
from typing import Literal
import pandas as pd
import duckdb as ddb
from pathlib import Path

from sqlalchemy import create_engine, text
from sqlalchemy_utils import create_database, database_exists


def connect_sql(
    db: str,
    host: str = "",
    user: str = "",
    pw: str = "",
    dbms: Literal["mssql", "sqlite", "postgres"] = "mssql",
    ensure_db_exists: bool = False,
) -> object:
    """
    Connects to any SQL database based on the given parameters.

    Args:
        db (str): The name of the database / sqlite-file.
        host (str, optional): The host name or IP address of the database server. Defaults to an empty string.
        user (str, optional): The username for authentication. Defaults to an empty string.
        pw (str, optional): The password for authentication. Defaults to an empty string.
        dbms (Literal['mssql', 'sqlite','postgres'], optional): The type of database management system. Defaults to 'mssql'.
        ensure_db_exists (bool, optional): Specifies whether to create the database if it does not exist. Defaults to False.

    Returns:
        Connection: The context object for the established database connection

    Remarks:
        - postgres
            - psycopg2-binary must be installed
            - example:
                con = my_connect_to_any_sql(
                    host='<instance>.postgres.database.azure.com',
                    db='eteste',
                    user='<user>n@<instance>',
                    pw='<password>',
                    dbms='postgres',
                    ensure_db_exists=False
                )

    """

    if dbms == "mssql":
        url = f"mssql://{user}:{pw}@{host}/{db}?driver=ODBC Driver 17 for SQL Server"
    elif dbms == "sqlite":
        url = f"sqlite:///{db}"
    elif dbms == "postgres":
        url = f"postgresql+psycopg2://{user}:{pw}@{host}/{db}"
    else:
        print("dbms not supported")
        return None

    engine = create_engine(
        url,  # * leave positional argument unnamed, since it was relabeled between versions..
        connect_args={"connect_timeout": 10},
    )

    # * ensure db exists
    if ensure_db_exists:
        if not database_exists(engine.url):
            create_database(engine.url)
            print(f"db {db} created")
        else:
            print(f"db {db} exists")

    # * now connect
    try:
        con = engine.connect()
    except Exception as e:
        print(e)

    return con



def load_sql_to_sqlite(
    con_source,
    file_db: str,
    list_tables: list[str],
    dict_meta: dict = None,
    top_n_rows: int=0,
    verbose: bool=True
) -> None:
    """
    Load SQL tables into a SQLite database.

    Args:
        con_source (object): The connection object to the source database.
        file_db (str): The path to the SQLite database file.
        list_tables (list[str]): A list of SQL tables to load. Each table can be specified as a string or a list of two strings, where the first string is the table name in the source database and the second string is the table name in the SQLite database (optional).
        dict_meta (dict, optional): A dictionary containing metadata to be written to the SQLite database. Defaults to None.
        top_n_rows (int, optional): The number of rows to load from each table. Defaults to 0.
        verbose (bool, optional): Whether to print progress messages. Defaults to True.

    Returns:
        None

    Raises:
        None

    Description:
        This function loads SQL tables from a source database into a SQLite database. If the SQLite database file already exists, the function exits. The function writes metadata to the SQLite database if a dictionary is provided. The function loads tables in batches of 10,000 rows and appends the loaded data to the corresponding table in the SQLite database.
    """
    # * check if db already exists
    if os.path.exists(file_db):
        print(f"❌ {file_db} already exists. exiting..")
        return

    con_sqlite = sqlite3.connect(file_db)
    batchsize = 10000

    # todo normalize on a set of terms like table_created_at, data_extracted_at, etc.
    # * write meta table if dict was given
    if dict_meta is not None:
        df_meta = pd.DataFrame.from_dict(dict_meta, orient="index").T
        df_meta.to_sql("_meta", con_sqlite, if_exists="replace", index=False)
    
    is_list_nested=all([isinstance(i,list) for i in list_tables])

    for item in list_tables:
        if is_list_nested:
            table_sql = item[0]
            table_friendly = item[1] if item[1] else item[0] if "." not in item[0] else item[0].split(".")[1]
        else:
            table_sql = item
            table_friendly = item if "." not in item else item.split(".")[1]
            
        # table_sql = item if not is_list_nested else item[0]
        # table_friendly = item if not is_list_nested else item[1] if item[1] else item[0] if "." not in item[0] else item[0].split(".")[1]
        if verbose:
            print(f"processing: {table_sql} -> {table_friendly}")

        top= f" top {top_n_rows}" if top_n_rows else ""
        qry = f"select{top} * from {table_sql}"

        proxy = (con_source
                .execution_options(stream_results=True)
                .execute(text(qry))
                )
        cols = list(proxy.keys())

        # todo pandas -> duckdb ??
        while 'batch not empty':
            batch = proxy.fetchmany(batchsize)
            df = pd.DataFrame(batch, columns=cols)
            df.to_sql(table_friendly, con_sqlite, if_exists="append", index=False)
            if not batch:
                break
        
    con_sqlite.close()
    return


def load_sqlite_to_parquet(
    file_sqlite: str,
    dir_local: str,
    fetch_views: bool = False,
    table_filter: str = "",
    overwrite: bool = False,
    verbose: bool = True,
    debug: bool = False,
) -> None:
    """
    Unpacks a SQLite database file into individual Parquet files for each table.
    This method uses duckdb engine.

    Args:
        file_sqlite (str): The path to the SQLite database file.
        dir_local (str): The directory where the Parquet files will be saved.
        fetch_views (bool, optional): Whether to include views in the unpacking. Defaults to False.
        tables_filter (str, optional): The optional filter to apply to the table names. Defaults to "". Is a pandas filter query, example: "table_name.str[0] == '_'"
        overwrite (bool, optional): Whether to overwrite existing files. Defaults to False.
        verbose (bool, optional): Whether to print the progress. Defaults to True.
        debug (bool, optional): Whether to debug. Defaults to False.

    Returns:
        None
    """
    
    if not os.path.exists(dir_local):
        os.makedirs(dir_local)

    # todo add support for views
    # views = ",'view'" if fetch_views else ""

    # * retrieve db name from file, this will be default name 
    db_name= os.path.basename(file_sqlite).split(".")[0]
    # * connect to db
    ddb.sql(f"attach if not exists '{file_sqlite}' as {db_name}; use {db_name};")
    # * retrieve all tables. this cant be filtered by database_name (weird effect)
    df_db=ddb.sql("SELECT * FROM duckdb_tables;").to_df()
    
    # * narrow down tables
    df_tbl = df_db.query(table_filter) if table_filter else df_db
    # * filter out other databases / connections / files
    df_tbl = df_tbl[df_tbl["database_name"] == db_name][["table_name"]]["table_name"].tolist()

    if debug:
        print("🧪 debugging 🧪")

    # * write tables in a loop
    for tbl in df_tbl:
        path = os.path.join(dir_local, f"{tbl}.parquet")
        exists = os.path.exists(path)

        if verbose or debug:
            if exists:
                if overwrite:
                    print(f"⏳ replacing: {path}")
                else:
                    print(f"💨 skipping: {path}")
            else:
                print(f"⏳ creating: {path}")

        if not debug and ((overwrite and exists) or (not exists)):
            # todo add top_n_rows
            qry=f"copy (select * from {tbl}) to '{path}'"
            ddb.sql(qry)

    return

def load_files_to_duckdb(
    dir: Path,
    ext: Literal["csv", "parquet"],
    list_files: list[str] = None,
    prefix: str = "",
    verbose: bool = False,
    debug: bool = False,
):
    """
    Load files from a given directory to a DuckDB database (['csv', 'parquet']).

    Args:
        dir (Path): The directory containing the files to unpack.
        ext (Literal["csv", "parquet"]): The file extension to unpack.
        list_files (list[str], optional): A list of files to unpack. Defaults to None.
        prefix (str, optional): A prefix to add to the unpacked files. Defaults to "".
        verbose (bool, optional): Whether to print loading messages. Defaults to False.
        debug (bool, optional): Whether to return a string instead of a DuckDB database. Defaults to False.

    Returns:
        Union[Tuple, str]: If debug is False, returns a tuple of DuckDB tables. If debug is True, returns a string of sorted file names.

    """
    files = set([os.path.basename(file).split(".")[0] for file in os.listdir(dir)])
    
    # * filter files if present
    if list_files is not None:
        files = files & set(list_files)
    
    # * exit or sort
    if files is None:
        return None
    else: files=sorted(files)

    items = []
    for file in files:
        if verbose:
            print(f"⏳ loading {file}")
        if not debug:
            if ext == "parquet":
                items.append(
                    ddb.read_parquet((dir / f"{file}.parquet").as_posix())
                )
            elif ext == "csv":
                items.append(
                    ddb.read_csv((dir / f"{file}.csv").as_posix(), header=True)
                )

    if not debug:
        # * unpacking the ddb files in tupel notation works
        # * tuple trick seems to not work on 1-item list
        out = items[0] if len(items)==1 else (*items,)
    else:
        files = [f"{prefix}{file}" for file in files]
        out = str(sorted(files)).replace("'", "").replace("[", "").replace("]", "")

    return out


import gnupg as gnu
import os

# * create object
gpg = gnu.GPG()

def _get_success_icon(success: bool) -> str:
    return "✅" if success else "❌"

def generate_key(
    name_email: str = "test@example.com",
    key_length: int = 2048,
    name_real: str = "slim shady",
    name_comment: str = "this is a test key",
    key_type: str = "RSA",
    passphrase: str = None,
) -> object:
    """
    Generate a PGP key with the given parameters.

    Args:
        name_email (str): The email address associated with the key. Defaults to "test@example.com".
        key_length (int): The length of the key in bits. Defaults to 2048.
        name_real (str): The real name associated with the key. Defaults to "slim shady".
        name_comment (str): A comment associated with the key. Defaults to "this is a test key".
        key_type (str): The type of key to generate. Defaults to "RSA".
        passphrase (str): The passphrase to protect the key. Defaults to PASSPHRASE_TEST.

    Returns:
        object: The generated PGP key object.
    """
    # * setup config
    input_data = gpg.gen_key_input(
        name_email=name_email,
        key_length=key_length,
        name_real=name_real,
        name_comment=name_comment,
        key_type=key_type,
        passphrase=passphrase,
    )

    # * generate raw key object
    key = gpg.gen_key(input_data)

    return key


def export_public_key(key: object, output_file_path: str = None) -> str:
    """
    A function to export a public key from an object id.

    Parameters:
        key (object): The object id from which the public key will be exported.
        output_file_path (str, optional): The file path where the public key will be saved. Defaults to None.

    Returns:
        str: The exported public key.
    """
    # * export public key from object id
    public_key = gpg.export_keys(key.__str__(), secret=False)

    # * print public key to have copyable text
    print(public_key)

    if output_file_path:
        with open(output_file_path, "w") as f:
            f.write(public_key)

    return public_key


def export_private_key(
    key: object,
    passphrase: str = None,
) -> str:
    """
    Export a private key from the given key object.

    Args:
        key (object): The key object to export the private key from.
        passphrase (str, optional): The passphrase to unlock the private key. Defaults to PASSPHRASE_TEST.

    Returns:
        str: The exported private key.

    """
    private_key = gpg.export_keys(key.__str__(), secret=True, passphrase=passphrase)

    return private_key


def import_key(key: str, key_file_path: str = None) -> object:
    """
    A function that imports a PGP key either from a string or a file path and returns the import results. Either public or private key can be imported.

    Parameters:
    key (str): The PGP key to import, either passed as a string or read from a file.
    key_file_path (str, optional): The file path to the PGP key file. If provided, the function reads the key from this file.

    Returns:
    object: The import results of the PGP key.
    """

    if key_file_path:
        with open(key_file_path, "r") as f:
            key = f.read()

    result = gpg.import_keys(key)
    print(result.results)
    return result.results


def encrypt(
    recipient_key_id_list: str | list[str],
    message: str = None,
    message_file_path: str = None,
) -> any:
    """
    Encrypts a message using PGP encryption for the specified recipient key IDs and optional passphrase.

    Args:
        message (str): The message to be encrypted.
        recipient_key_id_list (str | list[str]): The recipient key ID(s) for encryption.
        passphrase (str, optional): The passphrase for encryption. Defaults to PASSPHRASE_TEST.
        output_file_path (str, optional): The file path to save the encrypted message. Defaults to None.

    Returns:
        result: Result of the encryption.
    """
    if not message and not message_file_path:
        print("❌ no message or message file to encrypt")
        return None

    # * make str to list if only one key is provided
    if not isinstance(recipient_key_id_list, list):
        recipient_key_id_list = [recipient_key_id_list]

    if message_file_path:
        result = gpg.encrypt_file(
            message_file_path,
            recipients=recipient_key_id_list,
            output=message_file_path+".gpg",
            always_trust=True,  # ! if false, keys trust must be ultimate
        )
    else:
        result = gpg.encrypt(
            data=message,
            recipients=recipient_key_id_list,
            always_trust=True,  # ! if false, keys trust must be ultimate
        )

    # out = result.data.decode("utf-8")
    print(f"{_get_success_icon(result.ok)} encrypted message for recipient(s) {recipient_key_id_list}")
    return result


def decrypt(
    message: str = None,
    message_file_path: str = None,
    passphrase: str = None,
) -> any:
    """
    A function to decrypt a PGP message. Decrypts the provided message using a passphrase and returns the decrypted message as a string.
    Private key of recipient must be in keyring!

    Parameters:
    - message: str, the PGP message to decrypt
    - message_file_path: str, optional, the file path to the PGP message to decrypt
    - passphrase: str, the passphrase required for decryption

    Returns:
    - result: the result of the decryption

    hint: result.data.decode("utf-8")
    """

    if not message and not message_file_path:
        print("❌ no message or message file to decrypt")
        return None
    if message_file_path:
        # with open(message_file_path) as f:
        #     message = f.read()
        result = gpg.decrypt_file(
            fileobj_or_path=message_file_path,
            passphrase=passphrase,
            output=message_file_path[:-4],
            always_trust=True,
        )
    else:
        # * decrypt using secret key
        result = gpg.decrypt(
            message,
            passphrase=passphrase,
            always_trust=True,
        )

    # * decode byte stream to string
    print(f"{_get_success_icon(result.ok)} decrypted message")
    
    return result


def find_key(key_id: str, check_private: bool = False) -> str:
    """
    Finds a PGP key by its key ID. Either key id or fingerprint must be provided.

    Args:
        key_id (str): The key ID to search for. Must be at least 6 characters long.
        check_private (bool, optional): Whether to search for private keys. Defaults to False.

    Returns:
        str or None: The key ID if found, or None if not found.

    Raises:
        None
    """
    if len(key_id) < 8:
        print("❌ key_id must be at least 6 characters long")
        return None

    keys = gpg.list_keys(check_private)
    key_ids = [key["keyid"] for key in keys]
    key_prints = [key["fingerprint"] for key in keys]

    key_ids_match = [item for item in key_ids if key_id in item]
    key_prints_match = [item for item in key_prints if key_id in item]

    if key_ids_match:
        print(f"✅ key found: {key_ids_match}")
        return key_ids_match
    elif key_prints_match:
        print(f"✅ key found: {key_prints_match}")
        return key_prints_match
    else:
        print("❌ key not found")
        return None
